/* This file is part of Chunky Monkey IRCD 
 *
 * Chunky Monkey IRCD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Chunky Monkey IRCD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Chunky Monkey IRCD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * $Id: server.c,v 1.18.4.15 2003/02/04 23:31:50 lloydy Exp $
 *
 */
#include "server.h"

/* server.c - routines related to server activites
 * mainly contains code to add/remove servers to the
 * hash table. */


/* Actual table */
static ServerInfo servers[SERV_HASH_SIZE];
/* Pointer to our structure */
static ServerInfo my_ptr = NULL;

/* Routines */
static int get_subscript_by_name(char *);

/* initialise hash table, and add us */
int server_hash_init(char *my_name, char *my_info)
{
	int i;
	ServerInfo my_serv = NULL;

	/* NULL out hash */
	for (i = 0; i < SERV_HASH_SIZE; i++) {
		servers[i] = NULL;
	}

	/* Attempt to alloc memory for our info */
	my_serv = ircd_malloc(sizeof(struct _serverinfo));


	/* Setup Struct */
	strncpy(my_serv->name, my_name, SERVNAME_MAX);
	strncpy(my_serv->info, my_info, SERVNAME_MAX);
	my_serv->fd = -1;
	my_serv->hops = 0;
	my_serv->via_ptr = NULL;

	/* Add to hash */
	if (!server_hash_add(my_serv)) {
		log_msg(LOG_BASIC,
			"Created my info, but failed to add it to the hash");
		return -1;
	}
	my_ptr = my_serv;

	return 1;

}


/* Add a server into the hash table */
int server_hash_add(ServerInfo server)
{
	int subscript, i;

	/* Get inital subscript */
	subscript = get_subscript_by_name(server->name);

	/* Find a free one, if this isn't free */
	for (i = 1; servers[subscript] != NULL; i = 2 * i) {
		/* Move along */
		subscript += i;
		/* Reduce if too large */
		if (subscript >= SERV_HASH_SIZE)
			subscript -= SERV_HASH_SIZE;

	}


	/* Add in */
	servers[subscript] = server;

	return 1;

}

/* Remove a table from the hash */
/* Don't free() it */
int server_hash_rem(char *name)
{
	int subscript, i;

	subscript = get_subscript_by_name(name);


	for (i = 1;
	     servers[subscript] != NULL
	     && lc_strncmp(name, servers[subscript]->name, -1) != 0;
	     i = 2 * i) {
		subscript += i;
		if (subscript >= SERV_HASH_SIZE)
			subscript -= SERV_HASH_SIZE;
	}

	if (servers[subscript] == NULL) {
		/* Couldn't find it */
		return EXIT_FAILURE;
	} else {
		servers[subscript] = NULL;
		return EXIT_SUCCESS;
	}
}

/* Locate and return a pointer to the server with a name which matches this one */
/* return NULL on failure */
ServerInfo server_locate(char *name)
{
	int subscript, i;

	subscript = get_subscript_by_name(name);

	for (i = 1;
	     servers[subscript] != NULL
	     && lc_strncmp(name, servers[subscript]->name, -1) != 0;
	     i = 2 * i) {
		subscript += i;
		if (subscript >= SERV_HASH_SIZE)
			subscript -= SERV_HASH_SIZE;
	}

	if (servers[subscript] == NULL) {
		return NULL;
	} else {
		return servers[subscript];
	}
}

/* Get a subscript for the server name */
static int get_subscript_by_name(char *name)
{
	int subscript, i, total = 0;

	for (i = 0; name[i] != '\0'; i++) {
		if (name[i] < 96) {
			total += 2 * (name[i] + 32);
		} else {
			total += 2 * name[i];
		}
	}

	subscript = total % SERV_HASH_SIZE;

	return subscript;
}

/* rest of these commands are used to inform servers of occurances on this server */
/* 
 * Inform servers connected directly to us 
 * Takes a char * param to show who sent it
 * Usually used for events occuring on this server 
 */
int servers_announce(char *origin, char *text)
{
	int i;

	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && (servers[i]->fd != -1)
		    && servers[i] != my_ptr) {
			/* send */
			send_origin_server(servers[i], origin, text);
		}
	}

	return 1;
}

/*
 * Inform servers connected to us
 * used for events not occuring on this server
 */
int servers_relay(ServerInfo source, char *origin, char *text)
{
	int i;

	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && servers[i]->fd != -1
		    && servers[i] != my_ptr && servers[i] != source) {
			/* Send */
			send_origin_server(servers[i], origin, text);
		}
	}

	return 1;
}

/*
* Introduce a particular server
*/
static int introduce_server(ServerInfo recip, long int rand_no,
			    ServerInfo intro)
{
	char buffer[BUF_SIZE];

	if (intro->rand_no == rand_no) {
		return 0;
	} else {
		if (intro->via_ptr->rand_no != rand_no) {
			introduce_server(recip, rand_no, intro->via_ptr);
		}
		sprintf(buffer, "SERVER %s %0d 0 :%s", intro->name,
			intro->hops + 1, intro->info);
		send_origin_server(recip, intro->via_ptr->name, buffer);
		intro->rand_no = rand_no;
		return 1;
	}
}

/* 
 * Dump servers to a newly registered server
 * Don't send the server we're sending it do 
 */
int servers_introduce(ServerInfo recip)
{
	int i;
	long int rand_no;

	rand_no = rand() % 1000 + 1;
	get_my_server()->rand_no = rand_no;

	if (recip->fd == -1) {
		return 0;
	}

	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && servers[i] != recip
		    && servers[i] != my_ptr) {
			/* Send them the info */
			if (servers[i]->rand_no != rand_no) {
				introduce_server(recip, rand_no,
						 servers[i]);
			}
		}
	}

	return 1;
}


/* 
* Generate squits for servers connected via ours 
* s_ptr is server which died
* any server with via_ptr==s_ptr, should be killed 
*/
int server_netsplit(ServerInfo src, ServerInfo s_ptr, char *reason,
		    char *org_server_name)
{
	int i;
	char buffer[BUF_SIZE];
	ServerInfo temp = NULL;
	ServerInfo my_serv = NULL;

	my_serv = get_my_server();

	ircd_sprintf(buffer, BUF_SIZE, "SQUIT %s :%s", s_ptr->name,
		     reason);
	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && servers[i] != s_ptr
		    && servers[i]->via_ptr == s_ptr
		    && servers[i] != my_ptr) {

			/* Check for clients below this one */
			server_netsplit(src, servers[i], reason,
					org_server_name);

			/* Report it to non-ts4 servers */
			server_report_netsplit(src, my_serv->name, buffer);

			/* Work out which clients are dead */
			clients_netsplit(src, servers[i], org_server_name);

			/* Remove from hash */
			temp = servers[i];
			server_hash_rem(servers[i]->name);
			temp->via_ptr->server_count--;
			servers[i] = NULL;

			/* Free() it */
			server_free(temp);
		}

	}

	return 1;
}

/*
 * Used to report the netsplit info
 * Only sends to non-ts4 servers
 */
int server_report_netsplit(ServerInfo src, char *origin, char *buffer)
{
	unsigned int i;
	ServerInfo my_serv = NULL;

	/* Get my server */
	my_serv = get_my_server();

	/* Loop through the server list */
	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && servers[i] != src
		    && servers[i]->fd != -1) {
			if (servers[i]->ts_version < 4) {
				send_origin_server(servers[i], origin,
						   buffer);
			}
		}
	}

	return 1;
}

/* 
* Send the LINKS replies to the client 
*/
int server_links(ClientInfo c_ptr, char *mask)
{
	int i;
	char buffer[BUF_SIZE];

	if (c_ptr->fd == -1)
		return 0;

	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] && check_string(servers[i]->name, mask)) {
			/* It isn't us, and it matches the mask */
			ircd_sprintf(buffer, BUF_SIZE, "%s %s :%ld %s",
				     servers[i]->name,
				     servers[i] ==
				     my_ptr ? servers[i]->
				     name : servers[i]->via_ptr->name,
				     servers[i]->hops, servers[i]->info);

			send_numeric(c_ptr, get_my_server(), 364, buffer);
		}
	}

	return 1;
}

/*
 * free() the server information
 */
int server_free(ServerInfo serv_free)
{
#ifdef USE_OPENSSL
	if (serv_free->enc_ctx) {
		free(serv_free->enc_ctx);
	}
#endif
	free(serv_free);

	return 0;
}


/* 
* Return pointer to our server struct in memory 
*/
ServerInfo get_my_server()
{
	return my_ptr;
}

/*
* Generate LUSERS for servers which match a mask passed
*/
int servers_global_counts(char *mask, long int *client_count,
			  long int *server_count)
{
	int i;

	/* Set counters to 0 */
	(*client_count) = 0;
	(*server_count) = 1;

	for (i = 0; i < SERV_HASH_SIZE; i++) {
		if (servers[i] != NULL
		    && check_string(servers[i]->name, mask) != 0) {
			(*client_count) += servers[i]->client_count;
			(*server_count) += servers[i]->server_count;
		}
	}

	return 1;
}

/*
* Init a server connection
*/
int server_init_link(TempInfo t_ptr)
{
	char buffer[BUF_SIZE];
	struct cline *cline = NULL;
#ifdef USE_OPENSSL
	EncryptInfo ctx = NULL;
#endif

	/* Get the password for this server */
	cline = get_cline(t_ptr->name);

#ifdef USE_OPENSSL
	/* Send our TS4 capab line */
	ctx = generate_key((unsigned char *) cline->key);

	if (ctx) {
		log_msg(LOG_BASIC,
			"Gone encrypted in ircd_init_server_connection");
		ircd_sprintf(buffer, BUF_SIZE, "CAPAB :ENCRYPT");
		send_temp(t_ptr, buffer);
		t_ptr->enc_ctx = ctx;
	}
#endif


	if (!cline->password) {
		/* Try to connect w/o pass */
	} else {
		/* Send it */
		ircd_sprintf(buffer, BUF_SIZE, "PASS %s %s%s %s %s",
			     cline->password, "0280", VERSION,
			     "CMIRCD|Nothing", "");
		send_temp(t_ptr, buffer);
	}

	ircd_sprintf(buffer, BUF_SIZE, "CAPAB :TS4");
	send_temp(t_ptr, buffer);

	/* Send our info */
	ircd_sprintf(buffer, BUF_SIZE, "SERVER %s 0 0 :%s",
		     get_my_server()->name, get_my_server()->info);
	send_temp(t_ptr, buffer);

	t_ptr->flags |= USR_SENTMYINFO;
	t_ptr->flags &= ~USR_CONNECT;
	t_ptr->flags |= USR_PASS;

	return 1;
}

/*
 * Run a non-blocking connect() when 
 * a resolver lookup has returned
 */
int server_init_connect(TempInfo temp, char *ip)
{
	struct sockaddr_in sin_addr;

	log_msg(LOG_DEBUG2, "Running Connect() on fd %d", temp->fd);
	log_msg(LOG_DEBUG2, "IP is %s", ip);

	/* Its a pending server connect */
	/* Connect on it */
	sin_addr.sin_family = AF_INET;
	/* Copy IP to numeric dest, and into server structure */
	inet_aton(ip, &(sin_addr.sin_addr));
	temp->connect_ip.s_addr = sin_addr.sin_addr.s_addr;
	sin_addr.sin_port = htons(temp->port);
	memset(&(sin_addr.sin_zero), 0, 8);

	if (connect
	    (temp->fd, (struct sockaddr *) &sin_addr,
	     sizeof(struct sockaddr_in)) == -1 && errno != EINPROGRESS) {
		/* Only reached if a real error occured */
		log_msg(LOG_DEBUG1,
			"connect() failed in server_init_connect()");
		return -1;
	}

	/* Remove pending should connect() flag */
	temp->flags &= ~USR_SHOULDCONN;

	/* Add flag for connect() waiting */
	temp->flags |= USR_CONNECT;

	return 0;
}
